/*
 * Runways for connect
 * (c) 2011 Meta Meta LLC
 * 
 * MIT Licensed, see LICENSE.md
 */

/* dependencies */

var _routes = {};
var _methods = [];

// set up our routes, parameters, and our defaults
function routeMap(path) 
{
    var map = {};
    matches = path.match(/\{.+?\}/);
    map.regex = path.replace(/\{.+?\}/g,"(.+?)");

    map.paramCount = 0;
    map.params = new Array();
    
    if (matches) 
    {
        if (matches.length == 1) map.regex = path.replace(/\/\{.+?\}/g,"(.*)");
        map.root = path.replace(/\/\{.*$/,'');
        matches.forEach(
            function(element) 
            {
                map.params[map.paramCount] = {};
                
                element = element.replace(/^\{(.+)\}$/,"$1");

                if (element.indexOf(':'))
                {
                    _e = element.split(':');
                    map.params[map.paramCount].name = _e[0];
                    map.params[map.paramCount].value = _e[1];
                } else 
                {
                    map.params[map.paramCount].name = element;
                    map.params[map.paramCount].value = false;   
                }
                map.paramCount++;
            });
    } else 
    {
        map.root = path;
    }

    return map;
}

// check methods for a resource existing
function badResource(method, path)
{
    var _method = method;
    var _path = path;
   
    othermethods = _methods.filter(function(m) { if (m != _method) return true });

    non_method_routes = [];
    non_method_routes = othermethods.filter(function(m) { if (find(m,_path)) return true });

    // the root path was found for a different method
    if (non_method_routes.length > 0) return 405;

    // the root path was not found
    return 404;
}

// check the method specified for a route using the rooturl
function findByRoot(method,rootpath)
{
    var _method = method;
    var _rootpath = rooturl;

    if (_routes[_method] == undefined) return false;

    match = _routes[_method].filter(function(route) { if (route.map.root = _rootpath) return true; });

    return match;
}

// Oh yes, it's the joy of .filter!
function find(method, url) {
    
    var _method = method.toLowerCase();
    var _path = url;

    //SILLY! No routes!
    if (_routes[_method] == undefined) return false;    

    var match = [];
    
    _methods.push(_method);

    match = _routes[_method].filter(function(route) { if (_path.match('^' + route.map.regex + '$')) return true; });

    if (match.length == 0) return false;

    // Best match is what we want
    if (match.length > 1) 
    {
        _match = undefined;
        match.forEach(function(_r) { if ((!_match.map.regex.length) || (_r.map.regex.length > _match.map.regex.length)) _match = _r });
    }

    // reassign to itself, there can be only one
    match = match[0];

    // Okay, let's not get too clever - we need to parse the _path
    // and assign the appropriate matches to the parameters
    match.params = new Array();
    _newpath = _path.replace(match.map.root,'');
    _newpath = _newpath.replace(/^\//,'');

    var pathParams = _newpath.split('/');
    if (!_newpath.indexOf('/')) pathParams = [_newpath];
  
    if (_newpath.length > 0) 
    {
        // populate match.params
        pathParams.forEach(
            function(param,index) 
            { 
                try {
                    if ((match.map.params[index] != undefined) && (match.map.params[index].name != undefined))
                    { 
                        _pname = match.map.params[index].name;
                    } else 
                    {
                        _pname = index;
                    }
                } catch (e) 
                {
                    console.log(e);
                }
                match.params.push( {name: _pname , value: param}) 
            });
    }

    //make sure the parameters are available for callbacks
    match.params.get = 
        function(n) 
        { 
            var _n = n;

            //we have no parameters set? return defaults - by id only
            if (!match.params.length) 
            {
                _p = match.map.params.filter(function(parameter) { if (parameter.name == _n) return true });
                if (_p[0].value) return _p[0].value;
                return false;
            }

            //otherwise, let's get the active value
            if (typeof _n == 'string')
            {
                _p = match.params.filter(function(parameter) { if (parameter.name == _n) return true });

                // Maybe we didn't have anything there, let's get the default - if there is one
                if (!_p) _p = match.map.params.filter(function(parameter) { if (parameter.name == _n) return true });
                if (_p[0].value) return _p[0].value;
                
                //No value!
                return false;
            } else if ((typeof _n == 'number') && (_n < match.params.length)) 
            {
                if (_p = match.params[_n].value) return _p;
                return false;
            } 

            return false;
        };    

    return match;    
}

/* the following is for adding routes */
exports.add = function(name,path,callback,methods)
{
    if (!path) throw new Error(name + ': route requires path');
    if (!callback) throw new Error(name + ':route requires callback');

    var _path = path;
    var _callback = callback;
    var _name = name;

    // Just one method, no need to force an array
    if (typeof methods == 'string') methods = [methods];

    // default method is always GET
    if ((methods == []) || (methods == undefined)) methods = ['get'];
   
    methods.forEach(
        function(httpMethod) 
        {
            httpMethod = httpMethod.toLowerCase();
            try 
            {
                if (find(httpMethod,_path)) 
                    throw new Error(_name + ": " + httpMethod + " " + _path + " is already defined");
                    
                if (_routes[httpMethod] == undefined) _routes[httpMethod] = new Array();
        
                route = { name: _name, map: routeMap(_path), callback: _callback };
    
                _routes[httpMethod].push(route);
            
            } catch (e) 
            {
                console.log("Could not add route: " + httpMethod + " " + _path);
                console.log(e);
            }

        });

    return true;
}

/* Serve our routes! */
exports.routes = function() 
{
    return function(req,res,next) 
    {
        try 
        {
            if(route = find(req.method,req.url))
            {
                res.statusCode = 200;
                res.end(route.callback(req,route));
                return true;
            } else 
            {
                statusCode = badResource(req.method,req.url);
                switch(statusCode)
                {
                    case 404:
                        res.statusCode = statusCode;
                        res.end("404 Not Found");
                        break;
                    case 405:
                        res.statusCode = statusCode;
                        res.end("405 Not Supported");
                        break;
                    default:
                        throw new Error("There was an error in the request, and an error in handling the error");
                }
                return false;
            }
        } catch (e) 
        {
            res.statusCode = 500;
            res.end(e.message);
            return false;
        }
        next(); 
    }
}

